#include <cogent.h>

#include "CwLadder.h"
#include "cw_ladder_draw.h"
#include "cw_ladder_parser_cj.h"
#include "../api/memtype.h"
#include <photon/PtList.h>

/* prototype declarations */
PtWidgetClass_t *CreateCwLadderClass (void);

/* widget class pointer - class structure, create function */
PtWidgetClassRef_t WCwLadder = { NULL, CreateCwLadderClass };
PtWidgetClassRef_t *CwLadder = &WCwLadder; 

//-----------------------------------------------------------------------------
// загрузка значений по умолчанию
//-----------------------------------------------------------------------------
static void cwladder_dflts (PtWidget_t *widget)
{

	CwLadderWidget_t	*ladder	= ( CwLadderWidget_t * ) widget;
//	PtBasicWidget_t		*basic 	= ( PtBasicWidget_t * ) widget;

	ladder->dev_name = strdup ("CJ1M");
	ladder->busbar = calloc (9, sizeof (busbar_t));
	if (ladder->busbar == NULL) {
		fprintf (stderr, "cwladder_dflts was unable to allocate memory for ladder->busbar\n");
		exit (EXIT_FAILURE);
	}	
	memset (ladder->busbar, 0, sizeof (busbar_t) * 9);
	
	if (!(ladder->myself = IP_TaskCreateMe (IP_GetChannelID(), "ladder", NULL, NULL, 0)))
	{
	    fprintf (stderr, "Could not initialize Cascade IPC subsystem\n");
	}
	ladder->hmsg = IP_MsgCreate (NULL, IP_MsgDefaultSize(), 0);

	ladder->mg.ul.x = 10;
	ladder->mg.ul.y = 10;
	ladder->mg.lr.x = 170;
	ladder->mg.lr.y = 10;
	ladder->dx = 80;
	ladder->dy = 150;
	ladder->chars_in_line = 10;

	ladder->dev_name = "CJ1M";

	ladder->n_busbar = 9;
	parser_parse (&ladder->busbar [0], "/proj4/Omron/UprVAB.cxr", 	"Управление ВАБ");
	parser_parse (&ladder->busbar [1], "/proj4/Omron/Mashzal.cxr", 	"Машзал");
	parser_parse (&ladder->busbar [2], "/proj4/Omron/NaPulte.cxr", 	"На пульте");
	parser_parse (&ladder->busbar [3], "/proj4/Omron/ShPT1.cxr", 	"ШПТ 1");
	parser_parse (&ladder->busbar [4], "/proj4/Omron/ShPT2.cxr", 	"ШПТ 2");
	parser_parse (&ladder->busbar [5], "/proj4/Omron/RShPT.cxr", 	"ШПТ Р");
	parser_parse (&ladder->busbar [6], "/proj4/Omron/Speed.cxr", 	"Скорость");
	parser_parse (&ladder->busbar [7], "/proj4/Omron/Avarii.cxr", 	"Аварии");
	parser_parse (&ladder->busbar [8], "/proj4/Omron/Predupr.cxr", 	"Предупреждения");
}

//-----------------------------------------------------------------------------
// чистка
//-----------------------------------------------------------------------------
static void cwladder_destroy (PtWidget_t *widget)
{
	CwLadderWidget_t	*ladder	= ( CwLadderWidget_t * ) widget;

	if (ladder->busbar) {
		free (ladder->busbar);
		ladder->busbar = NULL;
	}
}

//-----------------------------------------------------------------------------
// функция прорисовки всего widget'a
//-----------------------------------------------------------------------------
static void cwladder_draw (PtWidget_t *widget, PhTile_t *damage)
{
	CwLadderWidget_t	*ladder = ( CwLadderWidget_t * ) widget;
	PhRect_t			canvas;

	PgSetDrawMode (Pg_DRAWMODE_OPAQUE);
	PtSuperClassDraw (PtBasic, widget, damage );
	PtCalcCanvas (widget, &canvas);

	PtClipAdd (widget, &canvas);

	if (ladder->busbar) {
		scheme_draw (widget, &canvas);
	} else {
		PgSetStrokeColor (Pg_BLACK);
		PgDrawRect (&canvas, Pg_DRAW_STROKE);
	}

	PtClipRemove ();
}

//-----------------------------------------------------------------------------
// определение размеров виджета
//-----------------------------------------------------------------------------
static void cwladder_extent (PtWidget_t *widget)
{
	CwLadderWidget_t	*ladder = ( CwLadderWidget_t * ) widget;
	busbar_t			*busbar = &ladder->busbar [ladder->selected];
	rung_t				*rung;
	PhDim_t 			dim;
	int					i;
	
	dim.w = ladder->mg.ul.x;
	dim.h = ladder->mg.ul.y + ladder->dy; 		
	if (busbar) {				
		for (i = 0; i < busbar->n_rung; i++) {
			rung = &busbar->rung [i];
			dim.h += (rung->n_level) * ladder->dy;
			dim.w = max (dim.w, (rung->n_cont_line_pos + 1) * ladder->dx + ladder->mg.ul.x);
		}
	} 
	dim.w += ladder->mg.lr.x;
	dim.w = max (dim.w, 975);

	PtResizeCanvas (widget, &dim);
	PtSuperClassExtent (PtBasic, widget);
}

//-----------------------------------------------------------------------------
// обработка событий (RAW)
//-----------------------------------------------------------------------------
static int raw_clbk (PtWidget_t *widget, void *apinfo, PtCallbackInfo_t *cbinfo)
{
	CwLadderWidget_t	*ladder = (CwLadderWidget_t *) widget;

	PhPointerEvent_t *event_data;
	PhRect_t *r;
	busbar_t *busbar, *srch_busbar;
	rung_t *rung, *srch_rung;
	element_t *element, *srch_element;
	symbol_t *arg, *srch_arg;
	int i, j;
	int a, b, c;
	int cur_x, cur_y;
	PhRect_t rect;
	char str [512];
	char *str_ptr;
	static struct PxTransCtrl	*trans = NULL;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	r = PhGetRects (cbinfo->event);

	// нажата нопка мыши
	if (cbinfo->event->type == Ph_EV_BUT_PRESS) {
		event_data = (PhPointerEvent_t *) PhGetData (cbinfo->event);
		
		if (ladder->busbar && (ladder->selected < ladder->n_busbar)) {
			cur_x = ladder->mg.ul.x;
			cur_y = ladder->mg.ul.y;
			busbar = &ladder->busbar [ladder->selected];
			for (i= 0; i< busbar->n_rung; i++) {
				rung = &busbar->rung [i];
				for (j= 0; j< rung->n_element; j++) {
					element = &rung->element [j];
					rect.ul.x = cur_x + element->line_pos * ladder->dx;								
					rect.ul.y = cur_y + element->level * ladder->dy;								
					rect.lr.x = rect.ul.x + ladder->dx;
					rect.lr.y = rect.ul.y + ladder->dy;
					
					// если координаты указателя мыши находятся внутри области, занимаемой элементом
					if (PhRectIntersect (&rect, r) != 0) {
						// снять отметки со всех элементов всех схем
						for (a= 0; a< ladder->n_busbar; a++) {
							srch_busbar = &ladder->busbar [a];
							for (b= 0; b< srch_busbar->n_rung; b++) {
								srch_rung = &srch_busbar->rung [b];
								for (c= 0; c< srch_rung->n_element; c++) {
									srch_element = &srch_rung->element [c];
									srch_element-> selected = 0;
								}
							}
						}
						// отметить найденный элемент
						element->selected = 1;						
						
						// если у элемента есть аргумент (рассматривается только первый)
						if (element->n_arg > 0) {
							// отобразить адрес и название аргумента
							if (ladder->label_wgt) {
								if (trans == NULL) {
									if ((trans = PxTranslateSet (NULL, "windows-1251")) == NULL) {
							        	fprintf (stderr, "cwladder: unknown translation charset 'windows-1251'\n");
										exit (EXIT_FAILURE);
									}
							    }	
								str_ptr = str;
								arg = &element->arg [0];
								if (arg->type != -1) {
									switch (arg->type) {
										case CIO:			
											str_ptr += sprintf (str_ptr, "%d.%02d", arg->channel, arg->bitnum);
											break;
										case WR:			
											str_ptr += sprintf (str_ptr, "W %d.%02d", arg->channel, arg->bitnum);	
											break;
										case TIM:			
											str_ptr += sprintf (str_ptr, "T %04d", arg->channel);
											break;
										case CNT:			
											str_ptr += sprintf (str_ptr, "C %04d", arg->channel);	
											break;
										default:
									}
								}
								if (strlen (arg->name)) {
									str_ptr += sprintf (str_ptr, " - ");
								   	str_ptr += PxTranslateStateToUTF (trans, arg->name, strlen (arg->name), NULL, str_ptr, 256);
								}
								if (strlen (arg->comment)) {
									str_ptr += sprintf (str_ptr, " - ");
								   	str_ptr += PxTranslateStateToUTF (trans, arg->comment, 256, NULL, str_ptr, 256);
								}
	   							*str_ptr = '\0';
								PtSetResource (ladder->label_wgt, Pt_ARG_TEXT_STRING, str, 0);							
							}
							
							// найти все элементы с аргументом идентичным найденному
							if (ladder->list_wgt) {						
								PtListDeleteAllItems (ladder->list_wgt);		
								for	(a= 0; a< ladder->n_busbar; a++) {
									srch_busbar = &ladder->busbar [a];
									for (b= 0; b< srch_busbar->n_rung; b++) {
										srch_rung = &srch_busbar->rung [b];
										for (c= 0; c< srch_rung->n_element; c++) {
											srch_element = &srch_rung->element [c];
											if (srch_element->n_arg > 0) {
												srch_arg = &srch_element->arg [0];
												arg = &element->arg [0];
												if ((srch_arg->mem_type == arg->mem_type)&& 
													(srch_arg->type == arg->type)) {
													if (srch_arg->channel == arg->channel) {
														if (srch_arg->bitnum == arg->bitnum) {
															const char *pstr = &str [0]; 
															str_ptr = str;
															str_ptr += sprintf (str_ptr, "%s", srch_busbar->name);
															switch (srch_element->type) {
																case etContact:
																	str_ptr += sprintf (str_ptr, "\tВход");
																	break;	
																case etCoil:
																	str_ptr += sprintf (str_ptr, "\tВыход");
																	break;	
																case etInstr:
																	str_ptr += sprintf (str_ptr, "\tИнструкция");
																	break;	
																case etOther:
																	str_ptr += sprintf (str_ptr, "\t");
																	break;
																default:
																	str_ptr += sprintf (str_ptr, "\t");
															}
															str_ptr += sprintf (str_ptr, "\t%s", srch_element->mnemonic);
								   							*str_ptr = '\0';
															PtListAddItems (ladder->list_wgt, &pstr, 1, 0);	
															srch_element->selected = 1;
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}	
				cur_y += rung->n_level * ladder->dy;						
			}
		}			
		PtDamageWidget (widget);
		//printf ("click on x = %d, y = %d\n", r->ul.x, r->ul.y);
	}
	
	return (Pt_CONTINUE);
}

//-----------------------------------------------------------------------------
// выбор схемы
//-----------------------------------------------------------------------------
static void cwladder_select (PtWidget_t *widget, PtArg_t const *arg, PtResourceRec_t const *res_rec)
{
	CwLadderWidget_t	*ladder = ( CwLadderWidget_t * ) widget;

	if (arg->value < ladder->n_busbar) {
		ladder->selected = arg->value;
		PtExtentWidget (widget);
		PtDamageWidget (widget);
	}
}

//-----------------------------------------------------------------------------
// создание класса
//-----------------------------------------------------------------------------
PtWidgetClass_t *CreateCwLadderClass( void )
{
	static const PtRawCallback_t callback = {
		Ph_EV_BUT_PRESS,	
		raw_clbk,	NULL
	};

	// define our resources
	static PtResourceRec_t resources[] = {
		{Cw_LADDER_SELECT, cwladder_select, 0, 
			Pt_ARG_IS_NUMBER (CwLadderWidget_t, selected), 0},
		{Cw_LADDER_LABEL, Pt_CHANGE_INVISIBLE, 0, 
			Pt_ARG_IS_POINTER (CwLadderWidget_t, label_wgt), 0},
		{Cw_LADDER_LIST, Pt_CHANGE_INVISIBLE, 0, 
			Pt_ARG_IS_POINTER (CwLadderWidget_t, list_wgt), 0},
	};

	// set up our class member values
	static PtArg_t args[] = {
		{ Pt_SET_VERSION, 110},
		{ Pt_SET_STATE_LEN, sizeof (CwLadderWidget_t) },
		{ Pt_SET_DFLTS_F, (long)cwladder_dflts },		
		{ Pt_SET_DESTROY_F, (long)cwladder_destroy },
		{ Pt_SET_DRAW_F, (long)cwladder_draw },
		{ Pt_SET_EXTENT_F, (long)cwladder_extent }, 
		{ Pt_SET_RAW_CALLBACKS, (long)&callback, 1 },
		{ Pt_SET_FLAGS, 0, Pt_RECTANGULAR },		
		{ Pt_SET_NUM_RESOURCES, sizeof (resources)/sizeof (resources [0])},
		{ Pt_SET_RESOURCES, (long)resources, sizeof (resources)/sizeof (resources [0])},
	};

	// create the widget class
	return (CwLadder->wclass = PtCreateWidgetClass (
		PtBasic, 0, sizeof (args)/sizeof (args[0]), args));
}