/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <photon/PxProto.h>
#include <cogent.h>

#include "cw_ladder_draw.h"
#include "CwLadder.h"

#include "cw_ladder_parser_cj.h"
#include "cw_ladder_mnemonics.h"
#include "../api/memtype.h"


void Draw_contact (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder);
void Draw_spec_contact (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder);
void Draw_coil (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder);
void Draw_instr (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder);

int get_bool_value (symbol_t *arg, CwLadderWidget_t *ladder);
int get_int_value (symbol_t *arg, CwLadderWidget_t *ladder);

//-----------------------------------------------------------------------------
// установка шрифта
//-----------------------------------------------------------------------------
int cw_ladder_set_font (int font_size)
{
    char Helvetica [MAX_FONT_TAG];

    if(PfGenerateFontName("Helvetica", 0, font_size, Helvetica) == NULL) {
        perror("Unable to find font");
        return -1;
    } else {
        PgSetFont (Helvetica);
    }
    return 0;
}

//-----------------------------------------------------------------------------
// прорисовка контакта
//-----------------------------------------------------------------------------
void Draw_contact (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder)
{
	PhPoint_t	p [10];
	PhPoint_t	p_text;
	PhRect_t	r;
	PgColor_t	fill_color;
	PgColor_t	stroke_color;
	int			stroke_width;
	int 		w, h;
	int			i;
	int 		prev_active = 0;
    char		utf_str [BUFSIZ*2];
    char		tmp_str [BUFSIZ];
    char		*str_ptr;
    int			str_len;
    symbol_t	*arg;
//-----------------------------------------------------------------------------
	
	w = rect.lr.x - rect.ul.x; 
	h = rect.lr.y - rect.ul.y; 

//	
//			(p[8])				(p(2))	  (p[4])			(p[9])
//			  ||				  ||     //	||				  ||	
//			(p[0])======= (p[1])  ||   //   ||  (p[6])======(p[7])
//								  || //     ||
//								(p[3])	  (p[5])
//
	p[0].x = rect.ul.x - elem->prev_pos [0]*w;		p[0].y = rect.ul.y + 80*6/10;
	p[1].x = rect.ul.x + w*4/10;					p[1].y = p[0].y;
	p[2].x = rect.ul.x + w*4/10;					p[2].y = rect.ul.y + 80*5/10;
	p[3].x = p[2].x;								p[3].y = rect.ul.y + 80*7/10;	
	p[4].x = rect.ul.x + w*6/10;					p[4].y = p[2].y;
	p[5].x = p[4].x;								p[5].y = p[3].y;
	p[6].x = rect.ul.x + w*6/10;					p[6].y = p[0].y;
	p[7].x = rect.lr.x + elem->next_pos*w;			p[7].y = p[0].y;
	p[8].x = p[0].x;								p[8].y = p[0].y - h * elem->left_vert [0];
	p[9].x = p[7].x;								p[9].y = p[7].y - h * elem->right_vert;
	r.ul = p[2];									r.lr = p[5];

	fill_color = PgSetFillColor	(Pg_GREEN);
	stroke_color = PgSetStrokeColor (Pg_GREEN);
	stroke_width = PgSetStrokeWidth (5);

	if (elem->active) {
		PgDrawRect (&r, Pg_DRAW_FILL_STROKE);
	}
	for (i = 0; i < elem->n_prev [0]; i++) {
		if (elem->prev [0][i] < 0) prev_active = 1;
		else prev_active |= rung->element [elem->prev [0][i]].out_active;	
	}
	if (prev_active) {
		PgDrawLine (&p[0], &p[1]);					
		if (elem->left_vert [0]) PgDrawLine (&p[0], &p[8]);
		if (elem->active) {
			PgDrawLine (&p[6], &p[7]);					
			if (elem->right_vert) PgDrawLine (&p[7], &p[9]);
		}
	}
	elem->out_active = elem->active && prev_active;

	PgSetFillColor (fill_color);
	PgSetStrokeColor (Pg_BLACK);
	PgSetStrokeWidth (2);
		
	PgDrawLine (&p[0], &p[1]);					
	PgDrawLine (&p[6], &p[7]);					
	PgDrawLine (&p[2], &p[3]);					
	PgDrawLine (&p[4], &p[5]);					
	if (elem->inverted) 	PgDrawLine (&p[3], &p[4]);					
	if (!!elem->left_vert [0])  PgDrawLine (&p[0], &p[8]);
	if (!!elem->right_vert) PgDrawLine (&p[7], &p[9]);
	
    PgSetTextColor (Pg_BLACK);
    p_text.x = rect.ul.x + w/2;
	p_text.y = rect.ul.y + 80*2/10;

	arg = &elem->arg [0];
	if (arg->type != -1) {
		switch (arg->type) {
			case CIO:			
				sprintf (tmp_str, "%d.%02d", arg->channel, arg->bitnum);
				break;
			case WR:			
				sprintf (tmp_str, "W %d.%02d", arg->channel, arg->bitnum);	
				break;
			case TIM:			
				sprintf (tmp_str, "T %04d", arg->channel);
				break;
			case CNT:			
				sprintf (tmp_str, "C %04d", arg->channel);	
				break;
			default:
		}
		PgDrawText (tmp_str, strlen(tmp_str), &p_text, Pg_TEXT_CENTER |Pg_TEXT_TOP );
	}

	if (strlen (arg->name)) {
		p_text.y = rect.ul.y + 80 - 80*2/10;
		str_ptr = utf_str;
	   	str_ptr += PxTranslateStateToUTF (trans, arg->name, ladder->chars_in_line, NULL, utf_str, BUFSIZ*2);
	   	*str_ptr = '\0';
		
		PgSetUnderline (Pg_BLACK, Pg_TRANSPARENT, 0);
		PgDrawText (utf_str, strlen (utf_str), &p_text, Pg_TEXT_CENTER |Pg_TEXT_TOP );
		r.ul.x = rect.ul.x + 5;
		r.ul.y = rect.ul.y + 80;
		r.lr.x = rect.lr.x - 5;
		r.lr.y = rect.lr.y;

		str_ptr = utf_str;
		str_len = strlen (arg->comment);
		for (i = 0; i < str_len; i += ladder->chars_in_line) {
			while (	i < str_len && (
					arg->comment [i] == ' ' || 
					arg->comment [i] == ',' || 
					arg->comment [i] == '.')) i++;
		   	str_ptr += PxTranslateStateToUTF (trans, &arg->comment [i], 
		   										ladder->chars_in_line, NULL, str_ptr, ladder->chars_in_line*2);
		   	*str_ptr++ = '\n';
		}	
		*(str_ptr-1) = '\0';
		PgSetUnderline (Pg_TRANSPARENT, Pg_TRANSPARENT, 0);
		PgDrawMultiTextArea (utf_str, 0, &r, Pg_TEXT_LEFT, Pg_TEXT_LEFT|Pg_TEXT_TOP, 0);
	}
//-----------------------------------------------------------------------------
	PgSetStrokeColor (stroke_color);
	PgSetStrokeWidth (stroke_width);
}

//-----------------------------------------------------------------------------
// прорисовка специального контакта (сравнение и т.п.)
//-----------------------------------------------------------------------------
void Draw_spec_contact (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder)
{
	PhPoint_t	p [10];
	PhPoint_t	p_text;
	PhRect_t	r;
	PgColor_t	fill_color;
	PgColor_t	stroke_color;
	int			stroke_width;
	int 		w, h;
	int			i;
	int 		prev_active = 0;
	char 		str [BUFSIZ];
    char		utf_str [BUFSIZ*2];
//-----------------------------------------------------------------------------
	
	w = rect.lr.x - rect.ul.x; 
	h = rect.lr.y - rect.ul.y; 

//	
//			(p[8])				(p(2))	  (p[4])			(p[9])
//			  ||				  ||     	||				  ||	
//			(p[0])======= (p[1])  ||        ||  (p[6])======(p[7])
//								  ||        ||
//								(p[3])	  (p[5])
//
	p[0].x = rect.ul.x - elem->prev_pos [0]*w;		p[0].y = rect.ul.y + 80*6/10;
	p[1].x = rect.ul.x + w*2/10;					p[1].y = p[0].y;
	p[2].x = rect.ul.x + w*2/10;					p[2].y = rect.ul.y + 80*5/10;
	p[3].x = p[2].x;								p[3].y = rect.ul.y + 80*7/10;	
	p[4].x = rect.ul.x + w*8/10;					p[4].y = p[2].y;
	p[5].x = p[4].x;								p[5].y = p[3].y;
	p[6].x = rect.ul.x + w*8/10;					p[6].y = p[0].y;
	p[7].x = rect.lr.x + elem->next_pos*w;			p[7].y = p[0].y;
	p[8].x = p[0].x;								p[8].y = p[0].y - h * elem->left_vert [0];
	p[9].x = p[7].x;								p[9].y = p[7].y - h * elem->right_vert;
	r.ul = p[2];									r.lr = p[5];

	fill_color = PgSetFillColor	(Pg_GREEN);
	stroke_color = PgSetStrokeColor (Pg_GREEN);
	stroke_width = PgSetStrokeWidth (5);

	if (elem->active) {
		PgDrawRect (&r, Pg_DRAW_FILL_STROKE);
	}
	for (i = 0; i < elem->n_prev [0]; i++) {
		if (elem->prev [0][i] < 0) prev_active = 1;
		else prev_active |= rung->element [elem->prev [0][i]].out_active;	
	}
	if (prev_active) {
		PgDrawLine (&p[0], &p[1]);					
		if (elem->left_vert [0]) PgDrawLine (&p[0], &p[8]);
		if (elem->active) {
			PgDrawLine (&p[6], &p[7]);					
			if (elem->right_vert) PgDrawLine (&p[7], &p[9]);
		}
	}
	elem->out_active = elem->active && prev_active;

	PgSetFillColor (fill_color);
	PgSetStrokeColor (Pg_BLACK);
	PgSetStrokeWidth (2);
		
	PgDrawLine (&p[0], &p[1]);					
	PgDrawLine (&p[6], &p[7]);					
	PgDrawLine (&p[2], &p[3]);					
	PgDrawLine (&p[4], &p[5]);					
	if (!!elem->left_vert [0])  PgDrawLine (&p[0], &p[8]);
	if (!!elem->right_vert) PgDrawLine (&p[7], &p[9]);

    PgSetTextColor (Pg_BLACK);

	switch (elem->subtype) {
		case ST_EQU:
			PgDrawTextArea ("=", 1, &r, Pg_TEXT_CENTER |Pg_TEXT_MIDDLE );
			break;	
		case ST_LES:
			PgDrawTextArea ("<", 1, &r, Pg_TEXT_CENTER |Pg_TEXT_MIDDLE );
			break;	
		case ST_GRT:
			PgDrawTextArea (">", 1, &r, Pg_TEXT_CENTER |Pg_TEXT_MIDDLE );
			break;	
	}
	
    p_text.x = rect.ul.x + w/2;
	p_text.y = rect.ul.y + 80*1/10;
	sprintf (str, "%s = &%hd", elem->arg [0].text, get_int_value (&elem->arg [0], ladder));
   	PxTranslateStateToUTF (trans, str, BUFSIZ, NULL, utf_str, BUFSIZ*2);
	PgDrawText (utf_str, strlen(utf_str), &p_text, Pg_TEXT_CENTER |Pg_TEXT_TOP );
	p_text.y = rect.ul.y + 80*3/10;
	sprintf (str, "%s = &%hd", elem->arg [1].text, get_int_value (&elem->arg [1], ladder));
   	PxTranslateStateToUTF (trans, str, BUFSIZ, NULL, utf_str, BUFSIZ*2);
	PgDrawText (utf_str, strlen(utf_str), &p_text, Pg_TEXT_CENTER |Pg_TEXT_TOP );

//-----------------------------------------------------------------------------
	PgSetStrokeColor (stroke_color);
	PgSetStrokeWidth (stroke_width);
}

//-----------------------------------------------------------------------------
// прорисовка катушки
//-----------------------------------------------------------------------------
void Draw_coil (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder)
{
	PhPoint_t	p [5];
	PhRect_t	r;
	PhPoint_t	p_text;
	PgColor_t	fill_color;
	PgColor_t	stroke_color;
	int			stroke_width;
	int 		w, h;
	int			i;
	int 		prev_active = 0;
    char		utf_str [BUFSIZ*2];
    symbol_t	*arg;
//-----------------------------------------------------------------------------
	
	w = rect.lr.x - rect.ul.x; 
	h = rect.lr.y - rect.ul.y; 
	
//
//								  ==========  (p[3])
//			(p[4])				  ||    // ||
//			  []				  ||   //  ||
//			(p[0])======== (p[1]) ||  //   || 
//								  || //    ||
//								  ||//	   ||
//						   (p[2]) ==========
//
	p[0].x = rect.ul.x - elem->prev_pos [0]*w;	p[0].y = rect.ul.y + 80*6/10;
	p[1].x = rect.ul.x + w*4/10;				p[1].y = p[0].y;
	p[2].x = rect.ul.x + w*4/10;				p[2].y = rect.ul.y + 80*8/10;
	p[3].x = rect.ul.x + w*8/10;				p[3].y = rect.ul.y + 80*4/10;	
	p[4].x = p[0].x;							p[4].y = p[0].y - h * elem->left_vert [0];
	r.ul.x = rect.ul.x + w*4/10;				r.ul.y = rect.ul.y + 80*4/10;
	r.lr.x = rect.ul.x + w*6/10;				r.lr.y = rect.ul.y + 80*8/10;

	fill_color   = PgSetFillColor	(Pg_GREEN);
	stroke_color = PgSetStrokeColor (Pg_GREEN);
	stroke_width = PgSetStrokeWidth (5);

	if (elem->active) {
		PgDrawRect (&r, Pg_DRAW_FILL_STROKE);
	}
	for (i = 0; i < elem->n_prev [0]; i++) {
		if (elem->prev [0][i] < 0) prev_active = 1;
		else prev_active |= rung->element [elem->prev [0][i]].out_active;	
	}
	if (prev_active) {
		PgDrawLine (&p[0], &p[1]);					
		if (elem->left_vert [0]) PgDrawLine (&p[0], &p[4]);
	}
	elem->out_active = elem->active && prev_active;

	PgSetFillColor (fill_color);
	PgSetStrokeColor (Pg_BLACK);
	PgSetStrokeWidth (2);
		
	PgDrawLine (&p[0], &p[1]);					
	PgDrawRect (&r, Pg_DRAW_FILL_STROKE);
	if (elem->inverted) 	PgDrawLine (&p[2], &p[3]);					
	if (!!elem->left_vert [0]) PgDrawLine (&p[0], &p[4]);
	
    PgSetTextColor (Pg_BLACK);
	
	arg = &elem->arg [0];
	// Адрес переменной
    p_text.x = rect.ul.x + w/2;
	p_text.y = rect.ul.y + 80/10;
	if (arg->type != -1) {
		switch (arg->type) {
			case CIO:			
				sprintf (utf_str, "%d.%02d", arg->channel, arg->bitnum);
				break;
			case WR:			
				sprintf (utf_str, "W %d.%02d", arg->channel, arg->bitnum);	
				break;
			default:
		}
		PgDrawText (utf_str, strlen (utf_str), &p_text, Pg_TEXT_CENTER |Pg_TEXT_TOP );
	}
	
	// Имя переменной
	PgSetUnderline (Pg_BLACK, Pg_TRANSPARENT, 0);
    p_text.x = p [3].x + 5;
	p_text.y = p [3].y;
	PxTranslateStateToUTF (trans, arg->text, 32, NULL, utf_str, BUFSIZ*2);
	PgDrawText (utf_str, strlen (utf_str), &p_text, Pg_TEXT_LEFT |Pg_TEXT_TOP );

	// Комментарии
	PgSetUnderline (Pg_TRANSPARENT, Pg_TRANSPARENT, 0);
    p_text.x = p [3].x + 5;
	p_text.y = p [2].y;
	PxTranslateStateToUTF (trans, arg->comment, BUFSIZ, NULL, utf_str, BUFSIZ*2);
	PgDrawText (utf_str, strlen (utf_str), &p_text, Pg_TEXT_LEFT |Pg_TEXT_BOTTOM );

//-----------------------------------------------------------------------------
	PgSetStrokeColor (stroke_color);
	PgSetStrokeWidth (stroke_width);
}

//-----------------------------------------------------------------------------
// прорисовка блока инструкции
//-----------------------------------------------------------------------------
void Draw_instr (PhRect_t rect, rung_t *rung, element_t *elem, struct PxTransCtrl *trans, CwLadderWidget_t *ladder)
{
	PhPoint_t	p [6];
	PhRect_t	r1, r2;
	PhPoint_t	p_text, p_text_comment;
	PgColor_t	fill_color;
	PgColor_t	stroke_color;
	int			stroke_width;
	int 		w, h;
	int			i, j;
	int 		prev_active = 0;
	char		utf_str [BUFSIZ*2];
	symbol_t	*arg = NULL;
//-----------------------------------------------------------------------------
	
	w = rect.lr.x - rect.ul.x; 
	h = rect.lr.y - rect.ul.y; 
	
//
//								  =============
//  (p[5])    (p[2])			  || mnemonic ||
//	  []	    []				  =============		
//	  []	    []				  || text [0] ||
//	  []	  (p[0])====== (p[1]) ============= 
//	  []						  || text [1] ||
//	(p[3])================ (p[4]) =============
//								  || text [2] ||
//						          =============
//

	p[0].x = rect.ul.x - elem->prev_pos [0]*w;		p[0].y = rect.ul.y + 80*6/10;
	p[1].x = rect.ul.x + w*1/10;					p[1].y = p[0].y;
	p[2].x = p[0].x;								p[2].y = p[0].y - h * elem->left_vert [0];

	p[3].x = rect.ul.x - elem->prev_pos [1]*w;		p[3].y = rect.ul.y + 80*6/10 + 100*2/10;
	p[4].x = rect.ul.x + w*1/10;					p[4].y = p[3].y;
	p[5].x = p[3].x;								p[5].y = p[3].y - h * elem->left_vert [1] - 100*2/10;

	r2.ul.x = rect.ul.x + w*1/10;			r2.ul.y = rect.ul.y + 100/20;	
	r2.lr.x = rect.lr.x - w*1/10;			r2.lr.y = rect.ul.y + 100 - 100/20;
	r1.ul.x = rect.ul.x + w*1/10;			r1.ul.y = rect.ul.y + 100/20;
	r1.lr.x = rect.lr.x - w*1/10;			r1.lr.y = rect.ul.y + 80*3/10;
	

	fill_color   = PgSetFillColor	(Pg_GREEN);
	stroke_color = PgSetStrokeColor (Pg_GREEN);
	stroke_width = PgSetStrokeWidth (5);

	if (elem->active) {
		PgDrawRect (&r1, Pg_DRAW_FILL_STROKE);
	}
	for (j = 0; j < elem->n_inputs; j++) {
		for (i = 0; i < elem->n_prev [j]; i++) {
			if (elem->prev [j][i] < 0) prev_active = 1;
			else prev_active |= rung->element [elem->prev [j][i]].out_active;	
		}
		if (prev_active) {
			PgDrawLine (&p[3*j + 0], &p[3*j + 1]);					
			PgDrawLine (&p[3*j + 0], &p[3*j + 2]);
		}
	}

	PgSetFillColor (fill_color);
	PgSetStrokeColor (Pg_BLACK);
	PgSetStrokeWidth (2);
		
	PgDrawRect (&r1, Pg_DRAW_FILL_STROKE);
	PgDrawRect (&r2, Pg_DRAW_FILL_STROKE);

	for (j = 0; j < elem->n_inputs; j++) {
		PgDrawLine (&p[3*j+0], &p[3*j+1]);					
		PgDrawLine (&p[3*j+0], &p[3*j+2]);
	}

    PgSetTextColor (Pg_BLACK);
    p_text.x = rect.ul.x + w/2;
	p_text.y = rect.ul.y + 80/10;
	p_text_comment.y = p_text.y;
	p_text_comment.x = rect.lr.x; 
	
   	PxTranslateStateToUTF (trans, elem->mnemonic, 32, NULL, utf_str, BUFSIZ*2);
	PgDrawText (utf_str, strlen (utf_str), &p_text, Pg_TEXT_CENTER |Pg_TEXT_TOP );
	p_text.y += 24;
	p_text_comment.y = p_text.y;
	for (i = 0; i < elem->n_arg; i++) {
		arg = &elem->arg [i];
	   	PxTranslateStateToUTF (trans, arg->text, 32, NULL, utf_str, BUFSIZ*2);
		PgDrawText (utf_str, strlen (utf_str), &p_text, Pg_TEXT_CENTER |Pg_TEXT_TOP );
		switch (arg->subtype) {
			case ST_BCD:
				sprintf (utf_str, "= #%hX", get_int_value (arg, ladder));
				break;
			case ST_DEC:
				sprintf (utf_str, "= &%hd", get_int_value (arg, ladder));
				break;
			case ST_BOOL:
				switch (arg->mem_type) {
					case WR_BIT:	
						sprintf (utf_str, "W %d.%02d = %hd", arg->channel, arg->bitnum, get_bool_value (arg, ladder));
						break;
					case CIO_BIT:	
						sprintf (utf_str, "%4d.%02d = %hd", arg->channel, arg->bitnum, get_bool_value (arg, ladder));
						break;
					default:	
						sprintf (utf_str, "= %hd", get_bool_value (arg, ladder));
						break;
				}
				break;
			case ST_TIM:
				sprintf (utf_str, "PV = #%hX", get_int_value (arg, ladder));
				break;
			case ST_CNT:
				sprintf (utf_str, "PV = #%hX", get_int_value (arg, ladder));
				break;
		}
		if ((arg->type) != -1)
			PgDrawText (utf_str, strlen (utf_str), &p_text_comment, Pg_TEXT_LEFT |Pg_TEXT_TOP );
		p_text.y += 24;
		p_text_comment.y = p_text.y;
	}

	PgSetStrokeColor (stroke_color);
	PgSetStrokeWidth (stroke_width);
}

///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// определение активности (0 или 1) для элемента
//-----------------------------------------------------------------------------
int get_bool_value (symbol_t *arg, CwLadderWidget_t *ladder) 
{
	int out = 0;
	char buf [255];
	char *buf_ptr;
	ST_STATUS	status;
	PT_stCPOINT point;
	
	if (ladder->dev_name) {
		buf_ptr = buf;
		buf_ptr += sprintf (buf, "%s_", ladder->dev_name);
		switch (arg->type) {
			case CIO: buf_ptr += sprintf (buf_ptr, "CIO_WRD"); break;
			case WR:  buf_ptr += sprintf (buf_ptr, "WR_WRD");  break;
			case CNT: buf_ptr += sprintf (buf_ptr, "CNT_WRD"); break;
			case TIM: buf_ptr += sprintf (buf_ptr, "TIM_WRD"); break;
			default:  return 0;
		}		
		buf_ptr += sprintf (buf_ptr, "_%06d", arg->channel);
		
		memset (&point, 0, sizeof (point));
		point.name = buf;
		
		if (ladder->myself) {
			if ((status = DH_ReadExistingPoint (ladder->myself, &point, ladder->hmsg, NULL)) != ST_OK) {
//				fprintf (stderr, "Read point \"%s\"failed: %s\n", point.name, ST_StatusName (status));
			}		
			out = (point.value.i >> arg->bitnum)& 1;
		} else {
			out = 0;
		}
	}
	return out;
}

//-----------------------------------------------------------------------------
// определение целочисленного значения (2 байта) для элемента
//-----------------------------------------------------------------------------
int get_int_value (symbol_t *arg, CwLadderWidget_t *ladder) 
{
	int out = 0;
	char buf [255];
	char *buf_ptr;
	ST_STATUS	status;
	PT_stCPOINT point;
	
	switch (arg->text [0]) {
		// шестнадцатеричная константа
		case '#':
			sscanf (&arg->text [1], "%X", &out);
			break;
		// десятичная константа
		case '&':
			sscanf (&arg->text [1], "%d", &out);
			break;			
			
		default :		
			if (ladder->dev_name) {
				buf_ptr = buf;
				buf_ptr += sprintf (buf, "%s_", ladder->dev_name);
				switch (arg->type) {
					case CIO: buf_ptr += sprintf (buf_ptr, "CIO_WRD"); break;
					case WR:  buf_ptr += sprintf (buf_ptr, "WR_WRD");  break;
					case CNT: buf_ptr += sprintf (buf_ptr, "CNT_WRD"); break;
					case TIM: buf_ptr += sprintf (buf_ptr, "TIM_WRD"); break;
					default:  return 0;
				}		
				buf_ptr += sprintf (buf_ptr, "_%06d", arg->channel);
		
				memset (&point, 0, sizeof (point));
				point.name = buf;
		
				if (ladder->myself) {
					if ((status = DH_ReadExistingPoint (ladder->myself, &point, ladder->hmsg, NULL)) != ST_OK) {
//						fprintf (stderr, "Read point \"%s\"failed: %s\n", point.name, ST_StatusName (status));
					}					
					out = point.value.i;
				} else {
					out = 0;
				}
			}
	}
	return out;
}

//-----------------------------------------------------------------------------
// определение размеров элемента
//-----------------------------------------------------------------------------
void calc_extent (element_t *elem, PhDim_t *dim)
{
//	dim->w = DELTA_X * (elem->prev_pos [0] + 1 + elem->next_pos);
//	dim->h = DELTA_Y * (max (elem->left_vert [0], elem->right_vert) + 1);
}

//-----------------------------------------------------------------------------
// прорисовка последней загруженной схемы
//-----------------------------------------------------------------------------
void scheme_draw (PtWidget_t *widget, PhRect_t *canvas)  
{
	PhRect_t     		rect;
	int 				next_num;
   	int 				i, j;
   	int 				cur_x = 0, cur_y = 0;
	rung_t				*rung;
	element_t			*elem;
	CwLadderWidget_t 	*ladder = (CwLadderWidget_t *) widget;
	busbar_t 			*busbar = &ladder->busbar [ladder->selected];
	
	static struct PxTransCtrl	*trans = NULL;
	
    if (trans == NULL) {
	    if ((trans = PxTranslateSet (NULL, "windows-1251")) == NULL) {
    	    fprintf (stderr, "cwladder_draw: unknown translation charset 'windows-1251'\n");
			exit (EXIT_FAILURE);
		}
    }	

	cur_x = ladder->mg.ul.x;
	cur_y = ladder->mg.ul.y;
	next_num = 0;
	
	if (busbar) {				
		cw_ladder_set_font (8);
		for (i = 0; i < busbar->n_rung; i++) {
			rung = &busbar->rung [i];
			for (j = 0; j < rung->n_element; j++) {
				elem = &rung->element [j];			
				if ((i > 0)&& (j == 0)) 
					elem->left_vert [0] = busbar->rung [i-1].n_level;	
////				calc_extent (elem, &dim);
				rect.ul.x = canvas->ul.x + cur_x + elem->line_pos * ladder->dx;								
				rect.ul.y = canvas->ul.y + cur_y + elem->level * ladder->dy;								
				rect.lr.x = rect.ul.x + ladder->dx;
				rect.lr.y = rect.ul.y + ladder->dy;
				if (elem->selected) {
					PgColor_t old_stroke_color;
					PgColor_t old_fill_color;
					old_stroke_color = PgSetStrokeColor (Pg_RED);
					old_fill_color = PgSetFillColor (Pg_GREY);
					PgDrawRect (&rect, Pg_DRAW_FILL_STROKE);
					PgSetStrokeColor (old_stroke_color);
					PgSetFillColor (old_fill_color);
				}
////				rect_intersect = rect;
////				if (!PhRectIntersect (&rect_intersect, canvas)) continue;
				switch (elem->type) {
				case etContact:
					if (elem->subtype == ST_NONE) {
						elem->active = get_bool_value (&elem->arg [0], ladder) ^ elem->inverted;
						Draw_contact (rect, rung, elem, trans, ladder);
					} else {
						switch (elem->subtype) {
							case ST_EQU:
								if (get_int_value (&elem->arg [0], ladder) == get_int_value (&elem->arg [1], ladder))
									elem->active = 1;
								else elem->active = 0;
								break;
							case ST_LES:
								if (get_int_value (&elem->arg [0], ladder) < get_int_value (&elem->arg [1], ladder))
									elem->active = 1;
								else elem->active = 0;
								break;
							case ST_GRT:
								if (get_int_value (&elem->arg [0], ladder) > get_int_value (&elem->arg [1], ladder))
									elem->active = 1;
								else elem->active = 0;
								break;	
						}
						Draw_spec_contact (rect, rung, elem, trans, ladder);
					}
					break;
				case etCoil:
					elem->active = get_bool_value (&elem->arg [0], ladder) ^ elem->inverted;
					Draw_coil (rect, rung, elem, trans, ladder);
					break;
				case etInstr:
					if ((elem->n_arg == 1)&& ((elem->arg [0].subtype) == ST_BOOL)) {
						elem->active = get_bool_value (&elem->arg [0], ladder);
					} else {
						elem->active = 0;	
					}
					Draw_instr (rect, rung, elem, trans, ladder);
					break;
				default: break;
				}
			}
			cur_y += rung->n_level * ladder->dy;
		}
	}
}