
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "cw_ladder_file_io.h"
#include "cw_ladder_mnemonics.h"


//-----------------------------------------------------------------------------
// прочитать очередную строку из файла
//-----------------------------------------------------------------------------
void read_line (parser_handle_t *handle)
{
	if (!fgets (handle -> buffer, BUFSIZ, handle -> fp)) {
		handle -> state = ST_EOF;
	} else {
		handle -> buffer [strlen (handle -> buffer) - 1] = 0;
		handle -> lnum ++;
	}
	handle -> bufptr = handle -> buffer;	
}

//-----------------------------------------------------------------------------
// открыть файл парсера
//-----------------------------------------------------------------------------
parser_handle_t *parser_open (const char *fname)
{
	parser_handle_t *handle;
	FILE			*fp;
	char			*f;
	
	f 		= NULL;
	fp 		= NULL;
	handle 	= NULL;
	
	if (fname) {
		fp = fopen (fname, "r");
		if (fp) {
			f = strdup (fname);
		} else {
			perror ("Open_parser");
		}
	} 
	
	if (fp) {
		handle = malloc (sizeof (*handle));
		memset (handle, 0, sizeof (*handle));
		handle -> fp = fp;
		handle -> fname = f;
		handle -> state = ST_RESET;
		read_line (handle);
	}
	
	return handle;
}	

//-----------------------------------------------------------------------------
// закрыть файл парсера
//-----------------------------------------------------------------------------
void parser_close (parser_handle_t *handle)
{
	if (handle) {
		fclose (handle -> fp);
		memset (handle, 0, sizeof (*handle));
		free (handle);
	}
}

//-----------------------------------------------------------------------------
// чтение слова из буфера
//-----------------------------------------------------------------------------
char* get_next_token (parser_handle_t *h)
{
	int state;
	
	h->token = NULL;
	state = SKIP_WHITESPACE;
	while (state != DONE) {
		switch (state) {
			case SKIP_WHITESPACE:
				if ((*h->bufptr > 0) && isspace (*h->bufptr)) {	
					h->bufptr ++;
					break;
				}
				if ((!*h->bufptr) || (*h->bufptr == ';')) {
					read_line (h);
					if (h -> state == ST_EOF) {
						state = DONE;
					}
					break;
				}
				h->token = h->bufptr;
				state = SKIP_TO_WHITESPACE_OR_EOL;
				h->bufptr ++;
				break;
				
			case SKIP_TO_WHITESPACE_OR_EOL:
				if ((*h->bufptr == 0) || ((*h->bufptr > 0) && isspace (*h->bufptr))) {
					*h->bufptr = 0;
					h->bufptr ++;
					state = DONE;
					break;
				}
				h->bufptr ++;
				break;
		}
	}		
	return h->token;
}

//-----------------------------------------------------------------------------
// поиск подходящего символа
//-----------------------------------------------------------------------------
const token_t* find_token (char *token)
{
	int i;
	
	if (token) {
		for (i = 0; i < sizeof (tokens) / sizeof (token_t); i++) {
			if (!stricmp (tokens [i].string, token)) {
				return &tokens [i];
			}
		}
	}
	return NULL;
}

//-----------------------------------------------------------------------------
// прочитать строку из буфера до конца или до символа табуляции
//-----------------------------------------------------------------------------
void get_next_str (parser_handle_t *h)
{
	h -> token = NULL;
	while ((*h -> bufptr > 0) && isspace (*h -> bufptr)) {	
		h -> bufptr ++;
	}
	h -> token = h -> bufptr;
	while (!((*h -> bufptr == '\n') || (*h -> bufptr == '\t') || (*h -> bufptr == 0) ||
			(*h -> bufptr == 0x0D) || (*(h -> bufptr+1) == 0x0A))) {
		h -> bufptr ++;
	}
	*h -> bufptr = 0;
}