///////////////////////////////////////////////////////////////////////////////
////  (c) Chaichuk AU (chaichuk@mail.ru), 								   ////
////  2004, Chelyabinsk, NTC Privodnaya Tehnika (momentum@chel.surnet.ru)  ////
///////////////////////////////////////////////////////////////////////////////

#ifndef __CW_LADDER_MNEMONICS_H_
#define __CW_LADDER_MNEMONICS_H_

typedef enum tok_e {	
	TOK_UNKNOWN	= -2,
	TOK_EOF		= -1,
	TOK_LD		=  1,
	TOK_AND,
	TOK_OR,
	TOK_OUT,
	TOK_INSTR,
	TOK_ANDLD,
	TOK_ORLD,
	TOK_END_RUNG,
	TOK_BEG_LOC_SYM_SEC,
	TOK_END_LOC_SYM_SEC,
	TOK_BEG_GLOB_SYM_SEC,
	TOK_END_GLOB_SYM_SEC,
	TOK_BEG_MNEMONIC_SEC,
	TOK_END_MNEMONIC_SEC,	
	TOK_BOOL,
} tok_t;

typedef enum st_e {
	ST_NONE	= 0,
	
	ST_EQU,			// 	==
	ST_LES,			//	<
	ST_GRT,			//	>
	ST_NEQU,		//	!=
	ST_LSEQ,		//	<=
	ST_GTEQ,		//	>=

	ST_BCD,
	ST_DEC,
	ST_FLT,
	ST_BOOL,
	ST_TIM,
	ST_CNT,
} st_t;


typedef struct token_s {
	char	*string;		// Мнемоника в тексте программы	
	tok_t	type;			// Тип элемента, соответствующий мнемонике
	st_t	subtype;		// Подтип элемента
	int		inverted;		// Инвертирован элемент (1) или нет (0)
	int		n_input;		// Кол-во входов у элемента
	int		n_arg;			// Кол-во аргументов, следующих за мнемоникой (до 3-х)
	st_t	arg_type [3];	// Типы аргументов
} token_t;

// Основные мнемоники, встречающиеся в тексте программы
const static token_t tokens [] =
{
	{"<MNEMONIC>",		TOK_BEG_MNEMONIC_SEC,	ST_NONE,	0,	0,	0,	{} },
	{"</MNEMONIC>",		TOK_END_MNEMONIC_SEC,	ST_NONE,	0,	0,	0,	{} },
	{"<LOCALSYMBOL>",	TOK_BEG_LOC_SYM_SEC,	ST_NONE,	0,	0,	0,	{} },
	{"</LOCALSYMBOL>",	TOK_END_LOC_SYM_SEC,	ST_NONE,	0,	0,	0,	{} },
	{"<GLOBALSYMBOL>",	TOK_BEG_GLOB_SYM_SEC,	ST_NONE,	0,	0,	0,	{} },
	{"</GLOBALSYMBOL>",	TOK_END_GLOB_SYM_SEC,	ST_NONE,	0,	0,	0,	{} },
	{"BOOL",			TOK_BOOL,				ST_NONE,	0,	0,	0,	{} },
	{"^^^",				TOK_END_RUNG,			ST_NONE,	0,	0,	0,	{} },

	{"LD",				TOK_LD,					ST_NONE,	0,	1,	1,	{ST_BOOL} },
	{"LDNOT",			TOK_LD,					ST_NONE,	1,	1,	1,	{ST_BOOL} },
	{"LD=(300)",		TOK_LD,					ST_EQU,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"LD<(310)",		TOK_LD,					ST_LES,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"LD>(320)",		TOK_LD,					ST_GRT,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"LD<>(305)",		TOK_LD,					ST_NEQU,	0,	1,	2,	{ST_BCD, ST_BCD} },
	{"LD<=(315)",		TOK_LD,					ST_LSEQ,	0,	1,	2,	{ST_BCD, ST_BCD} },
	{"LD>=(325)",		TOK_LD,					ST_GTEQ,	0,	1,	2,	{ST_BCD, ST_BCD} },

	{"AND",				TOK_AND,				ST_NONE,	0,	1,	1,	{ST_BOOL} },
	{"ANDNOT",			TOK_AND,				ST_NONE,	1,	1,	1,	{ST_BOOL} },
	{"AND=(300)",		TOK_AND,				ST_EQU,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"AND<(310)",		TOK_AND,				ST_LES,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"AND>(320)",		TOK_AND,				ST_GRT,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"AND<>(305)",		TOK_AND,				ST_NEQU,	0,	1,	2,	{ST_BCD, ST_BCD} },
	{"AND<=(315)",		TOK_AND,				ST_LSEQ,	0,	1,	2,	{ST_BCD, ST_BCD} },
	{"AND>=(325)",		TOK_AND,				ST_GTEQ,	0,	1,	2,	{ST_BCD, ST_BCD} },

	{"OR",				TOK_OR,					ST_NONE,	0,	1,	1,	{ST_BOOL} },
	{"ORNOT",			TOK_OR,					ST_NONE,	1,	1,	1,	{ST_BOOL} },
	{"OR=(300)",		TOK_OR,					ST_EQU,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"OR<(310)",		TOK_OR,					ST_LES,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"OR>(320)",		TOK_OR,					ST_GRT,		0,	1,	2,	{ST_BCD, ST_BCD} },
	{"OR<>(305)",		TOK_OR,					ST_NEQU,	0,	1,	2,	{ST_BCD, ST_BCD} },
	{"OR<=(315)",		TOK_OR,					ST_LSEQ,	0,	1,	2,	{ST_BCD, ST_BCD} },
	{"OR>=(325)",		TOK_OR,					ST_GTEQ,	0,	1,	2,	{ST_BCD, ST_BCD} },

	{"OUT", 			TOK_OUT,				ST_NONE,	0,	1,	1,	{ST_BOOL} },
	{"OUTNOT", 			TOK_OUT,				ST_NONE,	1,	1,	1,	{ST_BOOL} },

	{"ANDLD",			TOK_ANDLD,				ST_NONE,	0,	2,	0,	{} },
	{"ORLD", 			TOK_ORLD,				ST_NONE,	0,	2,	0,	{} },

	{"++",				TOK_INSTR,				ST_NONE,	0,	1,	1,	{ST_BCD} },
	{"SET",				TOK_INSTR,				ST_NONE,	0,	1,	1,	{ST_BOOL} },
	{"RSET",			TOK_INSTR,				ST_NONE,	0,	1,	1,	{ST_BOOL} },
	{"DIFU(013)",		TOK_INSTR,				ST_NONE,	0,	1,	1,	{ST_BOOL} },
	{"KEEP(011)",		TOK_INSTR,				ST_NONE,	0,	2,	1,	{ST_BOOL} },
	{"TIM",				TOK_INSTR,				ST_NONE,	0,	1,	2,	{ST_TIM, ST_BCD} },
	{"CNT",				TOK_INSTR,				ST_NONE,	0,	2,	2,	{ST_CNT, ST_BCD} },
	{"MOV(021)",		TOK_INSTR,				ST_NONE,	0,	1,	2,	{ST_BCD, ST_BCD} },
};

#endif //__CW_LADDER_MNEMONICS_H_