///////////////////////////////////////////////////////////////////////////////
////  (c) Chaichuk AU (chaichuk@mail.ru), 								   ////
////  2004, Chelyabinsk, NTC Privodnaya Tehnika (momentum@chel.surnet.ru)  ////
///////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fnmatch.h>

#include "../api/memtype.h"
#include "cw_ladder_parser_cj.h"
#include "cw_ladder_mnemonics.h"
#include "cw_ladder_file_io.h"

static 		element_t* 			add_element (rung_t *rung);
static 		element_t* 			copy_element (rung_t *rung, element_t *elem);
static		rung_t* 			add_rung (busbar_t *busbar);
static 		void 				delete_last_rung (busbar_t *busbar);
static 		symbol_t* 			add_symbol (busbar_t *busbar); 
static		void				calc_dim (rung_t *rung);

static 		element_t*			ld_contact (rung_t *rung);
static 		element_t*			and_contact (rung_t *rung);
static 		element_t*			or_contact (rung_t *rung);
static 		element_t*			out_coil (rung_t *rung);
static 		element_t*			out_instr_1 (rung_t *rung);
static 		element_t*			out_instr_2 (busbar_t *busbar, rung_t **rung);
static 		rung_t* 			and_last_blocks (busbar_t *busbar);
static 		rung_t* 			or_last_blocks (busbar_t *busbar);
static		void				set_TR (rung_t *rung, int tr_num);
static		void				get_TR (rung_t *rung, int tr_num);

static		void				update_addr (element_t *elem, busbar_t *busbar);


///////////////////////////////////////////////////////////////////////////////
// Главная функция парсера
///////////////////////////////////////////////////////////////////////////////
void parser_parse (busbar_t *busbar, const char *fname, const char *name)
{
//	busbar_t 		*out    	= NULL;
	rung_t			*rung   	= NULL;
	element_t		*elem   	= NULL;	
	symbol_t		*sym		= NULL;
	symbol_t		*arg		= NULL;
	const token_t	*token		= NULL;
	
	int			i, j, k;
	int 		tok;

	char		tmp_str [BUFSIZ];
	int			tmp_int;

	parser_handle_t *handle;

	handle = parser_open (fname);

busbar->name = strdup (name);
busbar->status = PARSER_OK;
while (handle->state != ST_EOF) {
	token = find_token (get_next_token (handle));
	if (token) tok = token->type;
	else tok = TOK_UNKNOWN;

	if (busbar->status == PARSER_OK) {
		switch (tok) {
		case	TOK_BEG_GLOB_SYM_SEC:
			busbar->status = PARSER_PARSE_GLOB_SYM;
			break;
		case	TOK_BEG_LOC_SYM_SEC:
			busbar->status = PARSER_PARSE_LOC_SYM;
			break;
		case	TOK_BEG_MNEMONIC_SEC:
			busbar->status = PARSER_PARSE_MNEMONIC;
			break;
		}
		continue;
	}

//= Global and Local symbols ==================================================

	if ((busbar->status == PARSER_PARSE_LOC_SYM) ||
		(busbar->status == PARSER_PARSE_GLOB_SYM)) {
		switch (tok) {

		case	TOK_END_GLOB_SYM_SEC:
			if (busbar->status == PARSER_PARSE_GLOB_SYM)
				busbar->status = PARSER_OK;
			break;
		case	TOK_END_LOC_SYM_SEC:
			if (busbar->status == PARSER_PARSE_LOC_SYM)
				busbar->status = PARSER_OK;			
			break;
			
		case	TOK_BOOL:	
			sym = add_symbol (busbar);
			strcpy (sym->name, tmp_str);
			get_next_token (handle);
			strcpy (sym->text, handle->token);	
			if (isdigit (handle->token [0])) {
				sym->type = CIO;
				sym->mem_type = CIO_BIT;
				sscanf (&handle->token [0], "%d.%d", &sym->channel, &sym->bitnum);
			} else {
				switch (handle->token [0]) {
					case 'W':
						sym->type = WR;
						sym->mem_type = WR_BIT;
						sscanf (&handle->token [1], "%d.%d", &sym->channel, &sym->bitnum);
						break;
					case 'T':
						sym->type = TIM;
						sym->mem_type = TIM_BIT;
						sscanf (&handle->token [1], "%d", &sym->channel);
						break;
					case 'C':
						sym->type = CNT;
						sym->mem_type = CNT_BIT;
						sscanf (&handle->token [1], "%d", &sym->channel);
						break;
					default:
						sym->type = -1;
						sym->mem_type = -1;				
				}	
			}
			get_next_str (handle);
			strcpy (sym->comment, handle->token);	
			break;
		case	TOK_UNKNOWN:
			strcpy (tmp_str, handle->token);
		}
		continue;
	}

//= Mnemonics =================================================================

	if (busbar->status == PARSER_PARSE_MNEMONIC) {
		switch (tok) {

		//---------------------------------------------------------------------
		case	TOK_END_MNEMONIC_SEC:
			if (rung) calc_dim (rung);
			busbar->status = PARSER_OK;
			break;
		
		//---------------------------------------------------------------------
		case	TOK_LD:		
			if (token->subtype == ST_NONE) {
				get_next_token (handle);
				if (handle->token [0] == 'T' &&
					handle->token [1] == 'R' &&
					isdigit (handle->token [2])) {
					if (sscanf (&handle->token [2], "%d", &tmp_int) == 1 &&
						tmp_int >= 0 && tmp_int <= 15) {
						get_TR (rung, tmp_int);
						break;
					}
				}
			}			
			rung = add_rung (busbar);
			elem = ld_contact (rung);			
			elem->inverted = token->inverted;			
			elem->subtype = token->subtype;
			strcpy (elem->mnemonic, token->string);			
			for (i = 0; i < token->n_arg; i++) {
				arg = &elem->arg [elem->n_arg++];
				if (token->subtype != ST_NONE) get_next_token (handle);
				strcpy (arg->text, handle->token);	
				arg->subtype = token->arg_type [i];
			}
			break;

		//---------------------------------------------------------------------
		case	TOK_AND:
			elem = and_contact (rung);
			elem->inverted = token->inverted;			
			elem->subtype = token->subtype;
			strcpy (elem->mnemonic, token->string);			
			for (i = 0; i < token->n_arg; i++) {
				arg = &elem->arg [elem->n_arg++];
				get_next_token (handle);
				strcpy (arg->text, handle->token);	
				arg->subtype = token->arg_type [i];
			}
			break;
			
		//---------------------------------------------------------------------
		case	TOK_OR:
			elem = or_contact (rung);
			elem->inverted = token->inverted;			
			elem->subtype = token->subtype;
			strcpy (elem->mnemonic, token->string);			
			for (i = 0; i < token->n_arg; i++) {
				arg = &elem->arg [elem->n_arg++];
				get_next_token (handle);
				strcpy (arg->text, handle->token);	
				arg->subtype = token->arg_type [i];
			}
			break;

		//---------------------------------------------------------------------
		case	TOK_ANDLD:
			rung = and_last_blocks (busbar);
			break;

		//---------------------------------------------------------------------
		case	TOK_ORLD:	
			rung = or_last_blocks (busbar);
			break;

		//---------------------------------------------------------------------
		case	TOK_OUT:		 	// Дискретный выход
			strcpy (tmp_str, handle->token);
			get_next_token (handle);			
			if (handle->token [0] == 'T' &&
				handle->token [1] == 'R' &&
				isdigit (handle->token [2])) {
				if (sscanf (&handle->token [2], "%d", &tmp_int) == 1 &&
					tmp_int >= 0 && tmp_int <= 15) {
					set_TR (rung, tmp_int);
					break;
				}
			}
			elem = out_coil (rung);
			arg = &elem->arg [elem->n_arg++];
			strcpy (elem->mnemonic, tmp_str); //handle->token);			
			strcpy (arg->text, handle->token);
			arg->subtype = token->arg_type [0];
			break;

		//---------------------------------------------------------------------
		case TOK_INSTR:
			if (token->n_input == 2) elem = out_instr_2 (busbar, &rung);
			else elem = out_instr_1 (rung);
			strcpy (elem->mnemonic, handle->token);			
			for (i = 0; i < token->n_arg; i++) {
				arg = &elem->arg [elem->n_arg++];
				get_next_token (handle);				
				strcpy (arg->text, handle->token);
				arg->subtype = token->arg_type [i];
			}
			break;

		//---------------------------------------------------------------------
		case	TOK_END_RUNG: 	
			calc_dim (rung);
			break;

		//---------------------------------------------------------------------
		case	TOK_UNKNOWN:		// Неизвестная инструкция
			fprintf (stderr, "unknown mnemonic \"%s\"\n", handle->token);
		}
		continue;
	}
}
	
for (i = 0; i < busbar->n_rung; i++) {
	busbar->max_line_pos = max (busbar->max_line_pos, busbar->rung [i].n_cont_line_pos);
}	
	
for (i = 0; i < busbar->n_rung; i++) {
	rung = &busbar->rung [i];
	for (j = 0; j < rung->n_element; j++) {
		elem = &rung->element [j];				
		update_addr (elem, busbar);
		if ((elem->type == etCoil) || (elem->type == etInstr)) {
			if (elem->n_inputs < 1) elem->n_inputs = 1;
			for (k = 0; k < elem->n_inputs; k++) {
				elem->prev_pos [k] += busbar->max_line_pos - elem->line_pos;
			}
			elem->line_pos = busbar->max_line_pos;
		}
	}
}
//-----------------------------------------------------------------------------	
	parser_close (handle);
}

///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// удалить общую шину и все ее элементы
//-----------------------------------------------------------------------------
void delete_busbar (busbar_t *busbar)
{
	rung_t *rung; 	
	int i;

	if (busbar->n_rung > 0) {
		for (i = 0; i < busbar->n_rung; i++) {
			rung = &busbar->rung [i];
			if (rung->n_element > 0) {
				rung->n_element = 0;
				free (rung->element);
				rung->element = NULL;
			}	
		}	
		busbar->n_rung = 0;
		free (busbar->rung);
		busbar->rung = NULL;
	}	
	if (busbar->n_symbol > 0) {
		busbar->n_symbol = 0;
		free (busbar->symbol);
		busbar->symbol = NULL;
	}	
}	

//-----------------------------------------------------------------------------
// добавить цепь на обшую шину
//-----------------------------------------------------------------------------
static rung_t* add_rung (busbar_t *busbar)
{
	rung_t *out; 	
	
	out = realloc (busbar->rung, (busbar->n_rung + 1)*sizeof (*out));
	if (out == NULL) {
		fprintf (stderr, "*** [library %s %d]:  add_rung was unable to allocate more memory (%d bytes), errno %d (%s)\n", __FILE__, __LINE__, sizeof (*out), errno, strerror (errno));
		return (NULL);
	}	
	busbar->n_rung ++;
	busbar->rung = out;
	memset (&busbar->rung [busbar->n_rung-1], 0, sizeof (*out));
	return (&busbar->rung [busbar->n_rung-1]);
}	

//-----------------------------------------------------------------------------
// удалить последнюю цепь/блок с общей шины
//-----------------------------------------------------------------------------
static void delete_last_rung (busbar_t *busbar)
{
	rung_t *out, *rung; 	

	if (busbar->n_rung < 1) return;	
		
	rung = &busbar->rung [busbar->n_rung - 1];
	if (rung->element) free (rung->element);
	out = realloc (busbar->rung, (busbar->n_rung - 1)*sizeof (*out));
	busbar->n_rung --;
	if (busbar->n_rung == 0) busbar->rung = NULL;
	else busbar->rung = out;
}	

//-----------------------------------------------------------------------------
// добавить элемент в цепь/блок
//-----------------------------------------------------------------------------
static element_t* add_element (rung_t *rung)
{
	element_t *out;

	out = realloc (rung->element, (rung->n_element + 1)*sizeof (*out));
	if (out == NULL) {
		fprintf (stderr, "*** [library %s %d]:  add_element was unable to allocate more memory (%d bytes), errno %d (%s)\n", __FILE__, __LINE__, sizeof (*out), errno, strerror (errno));
		return (NULL);
	}	
	rung->n_element ++;
	rung->element = out;
	memset (&rung->element [rung->n_element-1], 0, sizeof (*out));
	return (&rung->element [rung->n_element-1]);
}

//-----------------------------------------------------------------------------
// копировать элемент в цепь/блок
//-----------------------------------------------------------------------------
static element_t* copy_element (rung_t *rung, element_t *elem)
{
	element_t *out;

	out = realloc (rung->element, (rung->n_element + 1)*sizeof (*out));
	if (out == NULL) {
		fprintf (stderr, "*** [library %s %d]:  copy_element was unable to allocate more memory (%d bytes), errno %d (%s)\n", __FILE__, __LINE__, sizeof (*out), errno, strerror (errno));
		return (NULL);
	}	
	rung->n_element ++;
	rung->element = out;
	memcpy (&rung->element [rung->n_element-1], elem, sizeof (*elem));
	return (&rung->element [rung->n_element-1]);
}

//-----------------------------------------------------------------------------
// добавить символ/комментраии на общую шину
//-----------------------------------------------------------------------------
static symbol_t* add_symbol (busbar_t *busbar) 
{
	symbol_t *out;
		
	out = realloc (busbar->symbol, (busbar->n_symbol + 1)*sizeof (*out));
	if (out == NULL) {
		fprintf (stderr, "*** [library %s %d]:  add_symbol was unable to allocate more memory (%d bytes), errno %d (%s)\n", __FILE__, __LINE__, sizeof (*out), errno, strerror (errno));
		return (NULL);
	}	
	busbar->n_symbol ++;
	busbar->symbol = out;
	memset (&busbar->symbol [busbar->n_symbol-1], 0, sizeof (*out));
	return (&busbar->symbol [busbar->n_symbol-1]);
}

//-----------------------------------------------------------------------------
// обновление габаритных размеров цепи/блока
//-----------------------------------------------------------------------------
static void	calc_dim (rung_t *rung) 
{
	int i;
	element_t *elem;
	
	rung->n_level = 0;
	rung->n_line_pos = 0;
	rung->n_cont_line_pos = 0;
	rung->n_cont_level = 0;
	rung->n_out_level = 0;
	
	for (i = 0; i < rung->n_element; i++) {
		elem = &rung->element [i];
		rung->n_level = max (rung->n_level, elem->level+1);				
		rung->n_line_pos = max (rung->n_line_pos, elem->line_pos+1);
		if (elem->type == etContact) {
			rung->n_cont_line_pos = max (rung->n_cont_line_pos, elem->line_pos+1);
			rung->n_cont_level = max (rung->n_cont_level, elem->level+1); 	
		} else
		if (elem->type == etCoil || elem->type == etInstr) {
			rung->n_out_level = max (rung->n_out_level, elem->level+1);
		}
	}
	elem = &rung->element [rung->last_in_top_line];
	rung->n_out_level = max (rung->n_out_level, elem->level);
}

//-----------------------------------------------------------------------------
// добавить контакт - первый в цепи
//-----------------------------------------------------------------------------
static element_t* ld_contact (rung_t *rung)
{
	element_t *elem; 

	elem = add_element (rung);
	elem->type = etContact;
	elem->line_pos = 0;
	elem->level = 0;		
	
	elem->n_inputs = 1;
	elem->prev [0][0] = -1;
	elem->n_prev [0] = 1;
	
	rung->n_out_level = 0;
	rung->n_cont_level = 1;
	rung->n_cont_line_pos = 1;		
	rung->last_in_top_line = rung->n_element - 1;
	return elem;	
}

//-----------------------------------------------------------------------------
// добавить контакт - последовательно с предыдущими
//-----------------------------------------------------------------------------
static element_t* and_contact (rung_t *rung)
{
	element_t *elem; 
	element_t *last;
	int i;

	elem = add_element (rung);
	last = &rung->element [rung->last_in_top_line];

	elem->type = etContact;
	elem->line_pos = rung->n_cont_line_pos;
	elem->level = rung->n_out_level;			
	
	elem->n_inputs = 1;
	elem->left_vert [0] = elem->level - last->level;
	elem->prev_pos [0] = 0;

	elem->prev [0][0] = rung->last_in_top_line;
	elem->n_prev [0] = 1;
	for (i = 0; i < last->n_paral; i++) {
		elem->prev [0][elem->n_prev [0]++] = last->paral [i];
	} 
	
	rung->n_cont_line_pos ++;		
	rung->n_cont_level = max (rung->n_cont_level, elem->level+1);
	rung->last_in_top_line = rung->n_element - 1;
	return elem;	
}

//-----------------------------------------------------------------------------
// добавить контакт - параллельно с предыдущими
//-----------------------------------------------------------------------------
static element_t* or_contact (rung_t *rung)
{
	element_t *elem; 
	element_t *last;

	elem = add_element (rung);
	last = &rung->element [rung->last_in_top_line];
	
	elem->type = etContact;
	elem->line_pos = 0;
	elem->level = max (rung->n_cont_level, rung->n_out_level);
	
	elem->n_inputs = 1;
	elem->left_vert [0] = elem->level;
	elem->prev_pos [0] = 0;
	
	elem->right_vert = elem->level - last->level;
	elem->next_pos = last->line_pos;

	elem->prev [0][0] = -1;
	elem->n_prev [0] = 1;
	
	last->paral [last->n_paral++] = rung -> n_element - 1;
	
	rung->n_cont_level ++;
	return elem;
}

//-----------------------------------------------------------------------------
// добавить катушку
//-----------------------------------------------------------------------------
static element_t* out_coil (rung_t *rung)
{
	element_t *elem; 
	element_t *last;
	int i;

	elem = add_element (rung);
	last = &rung->element [rung->last_in_top_line];

	elem->type = etCoil;
	elem->line_pos = last->line_pos + 1;
	elem->level = rung->n_out_level;			
	
	elem->n_inputs = 1;
	elem->left_vert [0] = elem->level - last->level;
	elem->prev_pos [0] = 0;

	elem->prev [0][0] = rung->last_in_top_line;
	elem->n_prev [0] = 1;
	for (i = 0; i < last->n_paral; i++) {
		elem->prev [0][elem->n_prev [0]++] = last->paral [i];
	} 
	
	rung->n_out_level ++;		
	return elem;	
}

//-----------------------------------------------------------------------------
// добавить блок инструкции с одним входом
//-----------------------------------------------------------------------------
static element_t* out_instr_1 (rung_t *rung)
{
	element_t *elem; 
	element_t *last;
	int i;

	elem = add_element (rung);
	last = &rung->element [rung->last_in_top_line];

	elem->type = etInstr;
	elem->line_pos = rung->n_cont_line_pos;
	elem->level = rung->n_out_level;			
	
	elem->n_inputs = 1;
	elem->left_vert [0] = elem->level - last->level;
	elem->prev_pos [0] = 0;

	elem->prev [0][0] = rung->last_in_top_line;
	elem->n_prev [0] = 1;
	for (i = 0; i < last->n_paral; i++) {
		elem->prev [0][elem->n_prev [0]++] = last->paral [i];
	} 
	
	rung->n_out_level ++;		
	return elem;	
}

//-----------------------------------------------------------------------------
// добавить блок инструкции с двумя входами (две цепи/блока)
//-----------------------------------------------------------------------------
static element_t* out_instr_2 (busbar_t *busbar, rung_t **rung)
{
	rung_t *prev;
	rung_t *curr;
	element_t *elem; 
	element_t *out; 
	element_t *curr_last;
	element_t *prev_last;
	int offset;
	int i, j, k;

	if (busbar->n_rung < 2) return NULL;
	curr = &busbar->rung [busbar->n_rung - 1];
	prev = &busbar->rung [busbar->n_rung - 2];
	offset = prev->n_element;

//- объединение двух последних цепей ------------------------------------------
	
	for (i = 0; i < curr->n_element; i++) {
		elem = copy_element (prev, &curr->element [i]);
		elem->level += prev->n_cont_level;
		if (elem->n_prev [0] == 1 && elem->prev [0][0] == -1) {
			elem->left_vert [0] = elem->level;
		} else {
			for (k = 0; k < elem->n_inputs; k++) {
				for (j = 0; j < elem->n_prev [k]; j++) {
					elem->prev [k][j] += offset;	
				}	
			}
		}
		for (j = 0; j < elem->n_paral; j++) {
			elem->paral [j] += offset;
		}
	}

	curr_last = &prev->element [curr->last_in_top_line + offset];
	prev_last = &prev->element [prev->last_in_top_line];	

	calc_dim (prev);
	
//-----------------------------------------------------------------------------

	if (prev_last->line_pos < curr_last->line_pos) {
		for (i = prev->last_in_top_line; i < offset; i++) {
			elem = &prev->element [i];
			if (elem->type == etCoil || elem->type == etInstr) {
				elem->line_pos += (curr_last->line_pos - prev_last->line_pos);
			}
		}
		prev_last->next_pos = curr_last->line_pos - prev_last->line_pos;
	}

	out = add_element (prev);
	out->type = etInstr;
	out->line_pos = prev->n_cont_line_pos;
	out->level = prev->n_out_level;			
	
	out->n_inputs = 2;
	out->left_vert [0] = out->level - prev_last->level;
	out->prev_pos [0] = 0;
	out->left_vert [1] = out->level - curr_last->level;
	out->prev_pos [1] = 0;

	out->prev [0][0] = prev->last_in_top_line;
	out->n_prev [0] = 1;
	for (i = 0; i < prev_last->n_paral; i++) {
		out->prev [0][out->n_prev [0]++] = prev_last->paral [i];
	} 
	out->prev [1][0] = curr->last_in_top_line + offset;
	out->n_prev [1] = 1;
	for (i = 0; i < curr_last->n_paral; i++) {
		out->prev [1][out->n_prev [1]++] = curr_last->paral [i];
	} 
	
	prev->n_out_level ++;
	curr_last = &prev->element [curr->last_in_top_line + offset];
	prev_last = &prev->element [prev->last_in_top_line];	
	
	if (out->level >= curr_last->level) {
		for (i = prev->last_in_top_line; i < offset; i++) {
			elem = &prev->element [i];
			if (elem->type == etCoil || elem->type == etInstr) {
				elem->line_pos ++;
			}
		}
		prev_last->next_pos ++;
		out->line_pos ++;
		out->prev_pos [1] ++;
	}
	
	prev->last_in_top_line = curr->last_in_top_line + offset;
	calc_dim (prev);
	
	delete_last_rung (busbar);
	*rung = &busbar->rung [busbar->n_rung - 1]; 
	return out;	
}

//-----------------------------------------------------------------------------
// объединить два последних блока/цепи - последовательно. последнии удалить
//-----------------------------------------------------------------------------
static rung_t* and_last_blocks (busbar_t *busbar)
{
	rung_t *curr;
	rung_t *prev;
	element_t *elem;
	element_t *prev_last;
	int offset;
	int i, j, k;

	if (busbar->n_rung < 2) return NULL;
	curr = &busbar->rung [busbar->n_rung - 1];
	prev = &busbar->rung [busbar->n_rung - 2];
	offset = prev->n_element;

	for (i = 0; i < curr->n_element; i++) {
		elem = copy_element (prev, &curr->element [i]);
		prev_last = &prev->element [prev->last_in_top_line];	
		if (elem->n_prev [0] == 1 && elem->prev [0][0] == -1) {
			elem->prev [0][0] = prev->last_in_top_line;
			for (j = 0; j < prev_last->n_paral; j++) {
				elem->prev [0][elem->n_prev [0]++] = prev_last->paral [j];	
			}	
		} else {
			for (k = 0; k < elem->n_inputs; k++) {
				for (j = 0; j < elem->n_prev [k]; j++) {
					elem->prev [k][j] += offset;	
				}
			}	
		}
		for (j = 0; j < elem->n_paral; j++) {
			elem->paral [j] += offset;
		}
		elem->level += prev_last->level;
		elem->line_pos += prev->n_cont_line_pos;
	}
	
	prev->last_in_top_line = curr->last_in_top_line + offset;
	calc_dim (prev);
	delete_last_rung (busbar);	
	return &busbar->rung [busbar->n_rung - 1];
}

//-----------------------------------------------------------------------------
// объединить два последних блока/цепи - параллельно. последнии удалить
//-----------------------------------------------------------------------------
static rung_t* or_last_blocks (busbar_t *busbar)
{
	rung_t *curr;
	rung_t *prev;
	element_t *elem;
	element_t *prev_last;
	int offset;
	int i, j, k;

	if (busbar->n_rung < 2) return NULL;
	curr = &busbar->rung [busbar->n_rung - 1];
	prev = &busbar->rung [busbar->n_rung - 2];
	offset = prev->n_element;

	for (i = 0; i < curr->n_element; i++) {
		elem = copy_element (prev, &curr->element [i]);
		elem->level += prev->n_cont_level;
		if (elem->n_prev [0] == 1 && elem->prev [0][0] == -1) {
			elem->left_vert [0] = elem->level;
		} else {
			for (k = 0; k < elem->n_inputs; k++) {
				for (j = 0; j < elem->n_prev [k]; j++) {
					elem->prev [k][j] += offset;	
				}	
			}
		}
		for (j = 0; j < elem->n_paral; j++) {
			elem->paral [j] += offset;
		}
	}

	elem = &prev->element [curr->last_in_top_line + offset];
	prev_last = &prev->element [prev->last_in_top_line];	
	prev_last->paral [prev_last->n_paral++] = curr->last_in_top_line + offset;
	for (i = 0; i < elem->n_paral; i++) {
		prev_last->paral [prev_last->n_paral++] = elem->paral [i];	
	}
	if ((elem->line_pos + elem->next_pos) < prev_last->line_pos) {
		elem->next_pos = prev_last->line_pos - (elem->line_pos + elem->next_pos);
	} else
	if ((elem->line_pos + elem->next_pos) > prev_last->line_pos) {
		prev_last->next_pos = (elem->line_pos + elem->next_pos) - prev_last->line_pos;
	}
	elem->right_vert = elem->level - prev_last->level;				
	
	calc_dim (prev);
	delete_last_rung (busbar);	
	return &busbar->rung [busbar->n_rung - 1];
}

//-----------------------------------------------------------------------------
// запомнить начало ветвления rung->TR [tr_num]
//-----------------------------------------------------------------------------
static void set_TR (rung_t *rung, int tr_num)
{
	if (tr_num < 0 || tr_num > 15) return;
	rung->TR [tr_num] = rung->last_in_top_line;	

//	fprintf (stderr, "SAVE TR%d -> last_in_top_line = %d\n", tr_num, rung->last_in_top_line);
}

//-----------------------------------------------------------------------------
// переход на начало ветвления rung->TR [tr_num]
//-----------------------------------------------------------------------------
static void get_TR (rung_t *rung, int tr_num)
{
	element_t *last;
	
	if (tr_num < 0 || tr_num > 15) return;
	rung->last_in_top_line = rung->TR [tr_num];
	last = &rung->element [rung->last_in_top_line];
	rung->n_cont_line_pos = last->line_pos + last->next_pos + 1;
	rung->n_out_level = max (rung->n_out_level, rung->n_cont_level);

//	fprintf (stderr, "LOAD TR%d -> last_in_top_line = %d\n", tr_num, rung->last_in_top_line);
}

//-----------------------------------------------------------------------------
// обновление адресов элемента
//-----------------------------------------------------------------------------
static void update_addr (element_t *elem, busbar_t *busbar)
{
	int i, j;
	symbol_t *arg = NULL;
	symbol_t *sym = NULL;
	
	for (i = 0; i < elem->n_arg; i++) {
		arg = &elem->arg [i];
		for (j = 0; j < busbar->n_symbol; j++) {
			sym = &busbar->symbol [j];
			if (!strcmp (arg->text, sym->name)) {
				arg->type = sym->type;
				arg->mem_type = sym->mem_type;
				arg->channel = sym->channel;
				arg->bitnum = sym->bitnum;
				strcpy (arg->name, sym->name);	
				strcpy (arg->comment, sym->comment);
				break;
			}
		}
		if (j == busbar->n_symbol && arg->type == 0) {
			arg->name [0] = '\0';
			arg->comment [0] = '\0';
			if (isdigit (arg->text [0])) {
				switch (arg->subtype) {
					case ST_NONE: //- CIO ------------------------------------- 
						arg->type = CIO;
						arg->mem_type = CIO_BIT;
						sscanf (&arg->text [0], "%d.%d", &arg->channel, &arg->bitnum);
						break;
					case ST_TIM: //- Номер таймера -­--------------------------
						arg->type = TIM;
						arg->mem_type = TIM_BIT;
						sscanf (&arg->text [0], "%d", &arg->channel);
						break;
					case ST_CNT: //- Номер счетчика ---------------------------
						arg->type = CNT;
						arg->mem_type = CNT_BIT;
						sscanf (&arg->text [0], "%d", &arg->channel);
						break;
				}
			} else {
				switch (arg->text [0]) {
					case 'W': //- WR ------------------------------------------
						arg->type = WR;
						arg->mem_type = WR_BIT;
						sscanf (&arg->text [1], "%d.%d", &arg->channel, &arg->bitnum);
						break;
					case 'T': //- TIM -----------------------------------------
						arg->type = TIM;
						arg->mem_type = TIM_BIT;
						sscanf (&arg->text [1], "%d", &arg->channel);
						break;
					case 'C': //- CNT -----------------------------------------
						arg->type = CNT;
						arg->mem_type = CNT_BIT;
						sscanf (&arg->text [1], "%d", &arg->channel);
						break;
					case 'P': //- Clock Pulses ---------------------------------
						arg->type = -1;				
						sym->mem_type = -1;				
						strcpy (arg->name, arg->text);	
						if (!strcmp(arg->text, "P_1min")) {
							strcpy (arg->comment, "Meandr 1min");
							break;
						}
						if (!strcmp(arg->text, "P_1s")) {
							strcpy (arg->comment, "Meandr 1s");
							break;
						}
						if (!strcmp(arg->text, "P_0_2s")) {
							strcpy (arg->comment, "Meandr 0,2s");
							break;
						}
						if (!strcmp(arg->text, "P_0_1s")) {
							strcpy (arg->comment, "Meandr 0,1s");
							break;
						}
						if (!strcmp(arg->text, "P_0_02s")) {
							strcpy (arg->comment, "Meandr 0,02s");
							break;
						}												
					default: //------------------------------------------------
						arg->type = -1;				
						sym->mem_type = -1;				
				}					
			}
		}			
	}
}